-------------------------------------------------------------------------------
-- LISTENER by Tammya-MoonGuard (2016)
-------------------------------------------------------------------------------

local Main = ListenerAddon
local L = Main.Locale
local SharedMedia = LibStub("LibSharedMedia-3.0")

local g_history_size = 50
local g_lineid = 1

-- chat history database
--Main.chat_history = {
	--[[ 
		[playername]
		{
			[max_messages] = previous message history
			{
				id = line id
				t = time received (unixtime)
				e = event type
				m = message
				r = message has been read
				s = sender
			}
		}
	]]
--}

Main.unread_entries = {}
 
-------------------------------------------------------------------------------		  
function Main:OnInitialize() 
	SLASH_LISTENER1 = "/listener"
	SLASH_LISTENER2 = "/lr"
end

-------------------------------------------------------------------------------
function Main:CleanChatHistory()
	ListenerChatHistory = ListenerChatHistory or {}
	if ListenerChatHistory.version ~= Main.version then
		-- old version, drop history
		ListenerChatHistory = { 
			version = Main.version;
			data    = {};
		}
		self.chat_history = ListenerChatHistory.data
		return
	end
	
	self.chat_history = ListenerChatHistory.data
	
	local time = time()
	local expiry = 60*30 -- 30 mins
	
	-- we want to find the highest lineid so we can continue
	-- (it should reset overnight!)
	local max_lineid = 0
	
	for playername,chat_table in pairs( ListenerChatHistory.data ) do
		local writeto = 1
		
		for i = 1, #chat_table do
			if chat_table[i] then
				chat_table[i].r = true
				
				if time > chat_table[i].t + expiry then
					chat_table[i] = nil
				else
					local a = chat_table[i]
					chat_table[i] = nil
					chat_table[writeto] = a
					max_lineid = math.max( a.id, max_lineid )
					writeto = writeto + 1
				end
			end
		end
		--[[
		for entrykey,entry in pairs( chat_table ) do
			-- clear keys that have a time past the expiry duration
			if time > entry.t + expiry then
				chat_table[entrykey] = nil
			else
				
			-- compute the maximum of the message ids
				max_lineid = math.max( entry.id, max_lineid )
				entry.r = true
			end
		end
		
		-- we want to remove empty spaces in the list
		-- but we'll use a sorting function to keep things simple
		table.sort( chat_table, function( a, b ) 
					
			if a == nil then return false end
			if b == nil then return true end
			return a.id < b.id 
		end )
		]]
		-- if the list is empty, delete the player from the history
		if #chat_table == 0 then
			self.chat_history[playername] = nil 
		end
	end
	 
	g_lineid = max_lineid + 1
end

function Main:CleanPlayerList()

	for k,v in pairs(self.player_list) do
		if not self.chat_history[k] then
			self.player_list[k] = nil
		end
	end
	
	-- add friends list
	for i = 1,GetNumFriends() do
		local name = GetFriendInfo( i )
		if name and name ~= "Unknown" then
			self.player_list[name] = true
		end
	end
end

function Main:StartTooltip( frame )
	-- Section the screen into 6 sextants and define the tooltip 
	-- anchor position based on which sextant the cursor is in.
	-- Code taken from WeakAuras.
	--
    local max_x = 768 * GetMonitorAspectRatio()
    local max_y = 768
    local x, y = GetCursorPosition()
	
    local horizontal = (x < (max_x/3) and "LEFT") or ((x >= (max_x/3) and x < ((max_x/3)*2)) and "") or "RIGHT"
    local tooltip_vertical = (y < (max_y/2) and "BOTTOM") or "TOP"
    local anchor_vertical = (y < (max_y/2) and "TOP") or "BOTTOM"
    GameTooltip:SetOwner( frame, "ANCHOR_NONE" )
    GameTooltip:SetPoint( tooltip_vertical..horizontal, frame, anchor_vertical..horizontal )
	GameTooltip:ClearLines()
	
end

-------------------------------------------------------------------------------
function Main:OnEnable()

	self:CleanChatHistory() 

	self:Setup()
	self:CreateDB()

	self.MinimapButton:OnLoad()
	
	self.player_list = self.db.char.player_list
	self:CleanPlayerList()
	 
	
	self:RegisterEvent( "CHAT_MSG_SAY", "OnChatMsg" )
	self:RegisterEvent( "CHAT_MSG_EMOTE", "OnChatMsg" )
	self:RegisterEvent( "CHAT_MSG_TEXT_EMOTE", "OnChatMsg" )
--	self:RegisterEvent( "CHAT_MSG_WHISPER", "OnChatMsg" )
--	self:RegisterEvent( "CHAT_MSG_WHISPER_INFORM", "OnChatMsg" )
	self:RegisterEvent( "CHAT_MSG_PARTY", "OnChatMsg" )
	self:RegisterEvent( "CHAT_MSG_PARTY_LEADER", "OnChatMsg" )
	self:RegisterEvent( "CHAT_MSG_RAID", "OnChatMsg" )
	self:RegisterEvent( "CHAT_MSG_RAID_LEADER", "OnChatMsg" )
	self:RegisterEvent( "CHAT_MSG_RAID_WARNING", "OnChatMsg" )
	self:RegisterEvent( "CHAT_MSG_YELL", "OnChatMsg" )
	
	self:RegisterEvent( "UPDATE_MOUSEOVER_UNIT", "OnProbeUpdate" )
	self:RegisterEvent( "PLAYER_TARGET_CHANGED", "OnProbeUpdate" )
	
	self:RegisterEvent( "PLAYER_REGEN_DISABLED", "OnEnterCombat" )
	self:RegisterEvent( "PLAYER_REGEN_ENABLED", "OnLeaveCombat" )
	  
	self:Print( L["Version:"] .. " " .. self.version )
	   
	self:ApplyConfig()
	self:RefreshChat()
	
	self:Snoop_Setup()
	
	self:Init_OnEnabled()
end

-------------------------------------------------------------------------------
function Main:OnProbeUpdate()
	self:ProbePlayer()
end

-------------------------------------------------------------------------------
function Main:FlashClient()
	if Main.db.profile.flashclient then
		FlashClientIcon()
	end
end

local g_message_beep_cd = 0
function Main:PlayMessageBeep()
	if GetTime() < g_message_beep_cd + 0.05 then
		return
	end
	g_message_beep_cd = GetTime()
	
	PlaySoundFile( SharedMedia:Fetch( "sound", "ListenerBeep" ), "Master" )
	Main:FlashClient()
end

-------------------------------------------------------------------------------
-- Check if someone has used a text emote on us and then act accordingly.
-- 
-- @param msg Contents of text emote.
--
function Main:CheckPoke( msg, sender )
	if not Main.db.profile.sound.poke then return end
	
	sender = sender:gsub( "-.*", "" )
	
	local loc = GetLocale()
	if loc == "enUS" then
		if msg:find( " you" ) then
			if ListenerFrame:IsShown() and self.player_list[sender] then return end
			if UnitName('target') == sender then return end
			
			PlaySoundFile( SharedMedia:Fetch( "sound", "ListenerPoke" ), "Master" )
			Main:FlashClient()
		end
	end
end

-------------------------------------------------------------------------------
function Main:OnChatMsg( event, message, sender, language, _, _, _, _, _, _, _, _, guid )
	event = event:sub( 10 )
	if event == "TEXT_EMOTE" then
		if guid ~= UnitGUID( "player" ) then
			Main:CheckPoke( message, sender )
		end
	end
	self:AddChatHistory( sender, event, message, language, guid )
	self:Snoop_DoUpdate()
end

-------------------------------------------------------------------------------
local function GetTRPCharacterInfo( name )
	
	local char, realm = TRP3_API.utils.str.unitIDToInfo( name )
	if not realm then
		realm = TRP3_API.globals.player_realm_id
	end
	name = TRP3_API.utils.str.unitInfoToID( char, realm )
	
	if name == TRP3_API.globals.player_id then
		return TRP3_API.profile.getData("player");
	elseif TRP3_API.register.isUnitIDKnown( name ) then
		return TRP3_API.register.getUnitIDCurrentProfile( name ) or {};
	end
	return {};
end

-------------------------------------------------------------------------------
local function GetCharacterClassColor( guid )
	  
	local _, cls = GetPlayerInfoByGUID( guid )
	if cls and RAID_CLASS_COLORS[cls] then
		local c = RAID_CLASS_COLORS[cls]
		return ("|cff%.2x%.2x%.2x"):format(c.r*255, c.g*255, c.b*255)
	end
end

function Main:OnEnterCombat()
	if self.db.profile.combathide then
		self:HideFrame( true )
	end
end

function Main:OnLeaveCombat()
	if self.db.profile.combathide and not self.db.profile.frame.hidden then
		self:ShowFrame( true )
	end
end

-------------------------------------------------------------------------------
function Main:AddChatHistory( sender, event, message, language, guid )
	
	--local time = date("*t")
	--time = string.format( "[%02d:%02d]", time.hour, time.min );
	
	local original_sender = sender
	
	if message == "" then return end
	
	-- just strip the realm. collisions are -very- rare, and it just
	-- messes things up
	sender = sender:gsub( "-.*", "" )
	
	self.chat_history[sender] = self.chat_history[sender] or {}
	
	local isplayer = guid == UnitGUID( "player" )
	 
	local entry = {
		id = g_lineid;
		t = time();
		e = event;
		m = message;
		s = sender;
		g = guid;
	}
	
	if isplayer then
		entry.p = true -- is player
		entry.r = true -- read
	end
	
	if event == "YELL" then
		entry.r = true -- yells dont cause unread messages
	end
	
	g_lineid = g_lineid + 1
	
	table.insert( self.chat_history[sender], entry )
	
	while #self.chat_history[sender] > g_history_size do
		table.remove( self.chat_history[sender], 1 )
	end
	
	if not entry.r then
		table.insert( self.unread_entries, entry )
	end
	
	local add_message = true
	
	if entry.p then
		-- player is posting...
	--	if event == "SAY" or event == "EMOTE" or event == "TEXT_EMOTE" then
			-- player is publicly emoting, clear read status
			Main:MarkAllRead( true )
			--add_message = false
	--	end
	end
	
	if Main.db.profile.sound.target and guid == UnitGUID("target") and guid ~= UnitGUID("player") then
		Main:PlayMessageBeep()
	end
	
	if add_message and (self.player_list[ sender ] or guid == UnitGUID("player")) then
		Main:AddMessage( entry, true )
		Main:UpdateUnreadHighlight()
	end
end

-------------------------------------------------------------------------------
function SlashCmdList.LISTENER( msg )
	local args = {}
	
	for i in string.gmatch( msg, "%S+" ) do
		table.insert( args, i )
	end
	
	if args[1] == nil then
		Main:OpenConfig()
		return
	end
	
	if args[1] ~= nil then args[1] = string.lower( args[1] ) end
	
	if args[1] == "read" or args[1] == L["read"] then
		
	elseif args[1] == "add" or args[1] == L["add"] then
		if args[2] == nil then
			args[2] = UnitName( "target" )
		end
		if args[2] == nil then
			Main:Print( L["Specify name or target someone."] )
			return
		end
		
		Main:AddPlayer( args[2] )
		
	elseif args[1] == "remove" or args[1] == L["remove"] then
		
		if args[2] == nil then
			args[2] = UnitName( "target" )
		end
		if args[2] == nil then
			Main:Print( L["Specify name or target someone."] )
			return
		end
		
		Main:RemovePlayer( args[2] )
	elseif args[1] == "toggle" or args[1] == L["toggle"] then
		if args[2] == nil then
			args[2] = UnitName( "target" )
		end
		if args[2] == nil then
			Main:Print( L["Specify name or target someone."] )
			return
		end
		
		Main:TogglePlayer( args[2] )
		
	elseif args[1] == "clear" or args[1] == L["clear"] then
		Main:ClearAllPlayers()
		
	elseif args[1] == "list" or args[1] == L["list"] then
		Main:ListPlayers()
	elseif args[1] == "show" or args[1] == L["show"] then
		Main:ShowFrame()
	elseif args[1] == "hide" or args[1] == L["hide"] then
		Main:HideFrame()
	end
	  
end

local function FixupName( name )
	name = name:lower()
	
	-- (utf8 friendly) capitalize first character
	name = name:gsub("^[%z\1-\127\194-\244][\128-\191]*", string.upper)
	return name
end

-------------------------------------------------------------------------------
-- Add player to listening list.
--
-- @param name   Name of player.
-- @param silent Don't print confirmation/failure in chat.
-------------------------------------------------------------------------------
function Main:AddPlayer( name, silent )
	name = FixupName( name )
	
	if name == UnitName( "player" ) then return end
	
	if not self.player_list[name] then
		self.player_list[name] = true
		if not silent then self:Print( L["Added: "] .. name ) end
		Main:RefreshChat()
	else
		if not silent then self:Print( L["Already listening: "] .. name ) end
	end
	
	self:ProbePlayer()
end

-------------------------------------------------------------------------------
function Main:RemovePlayer( name, silent )
	name = FixupName( name )
	
	if self.player_list[name] then
		self.player_list[name] = nil
		if not silent then self:Print( L["Removed: "] .. name ) end
		Main:RefreshChat()
	else
		if not silent then self:Print( L["Not listening: "] .. name ) end
	end
	self:ProbePlayer()
end

-------------------------------------------------------------------------------
function Main:TogglePlayer( name, silent )
	name = FixupName( name )
	
	if self.player_list[name] then
		self:RemovePlayer( name, silent )
	else
		self:AddPlayer( name, silent )
	end
	self:ProbePlayer()
end

-------------------------------------------------------------------------------
function Main:MarkAllRead( forcerefresh )


	if #self.unread_entries == 0 and not forcerefresh then return end
	
	ListenerFrameBarRead.tex:SetColorTexture( 0.95,0.90,0.13,0.2 )
	local new_list = {}
	
	local time = time()
	for k,v in pairs( self.unread_entries ) do
		if time < v.t + 4 then
			-- this message is pretty fresh, its probably not quite read yet!
			table.insert( new_list, v )
		else
			v.r = true 
		end
	end
	self.unread_entries = new_list
	Main:CheckUnread()
	Main:UpdateUnreadHighlight()
end

-------------------------------------------------------------------------------
function Main:ClearAllPlayers()
	wipe( self.player_list )
	self:Print( L["Cleared all players."] )
	Main:RefreshChat()
end

-------------------------------------------------------------------------------
function Main:ListPlayers()
	local list = ""
	for k,v in pairs( self.player_list ) do
		list = list .. k .. " "
	end
	self:Print( L["Player list: "] .. list )
end

-------------------------------------------------------------------------------
function Main:Print( text )
	text = tostring( text )
	local prefix = "|cff9e5aea<Listener>|r "
	print( prefix .. text )
end
